# Copyright 2022 The JaxLinOp Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from dataclasses import dataclass

from beartype.typing import (
    Any,
    Tuple,
    Union,
)
import jax.numpy as jnp
from jaxtyping import Float

from gpjax.linops.diagonal_linear_operator import DiagonalLinearOperator
from gpjax.linops.linear_operator import LinearOperator
from gpjax.linops.utils import (
    check_shapes_match,
    default_dtype,
    to_linear_operator,
)
from gpjax.typing import (
    Array,
    ScalarFloat,
)


def _check_size(shape: Any) -> None:
    if not isinstance(shape, tuple):
        raise ValueError(
            f"`shape` must be a a tuple, but `type(shape) = {type(shape)}`."
        )

    if len(shape) != 2:
        raise ValueError(f"`shape` must be a 2-tuple, but `shape = {shape}`.")


# TODO: Generalise to non-square matrices.
@dataclass
class ZeroLinearOperator(LinearOperator):
    """Zero linear operator."""

    def __init__(self, shape: Tuple[int, ...], dtype: jnp.dtype = None) -> None:
        _check_size(shape)

        if dtype is None:
            dtype = default_dtype()

        self.shape = shape
        self.dtype = dtype

    def diagonal(self) -> Float[Array, " N"]:
        """
        Diagonal of the covariance operator.

        Returns
        -------
            Float[Array, " N"]: The diagonal of the covariance operator.
        """
        return jnp.zeros(self.shape[0])

    def __add__(
        self, other: Union[Float[Array, "N N"], LinearOperator]
    ) -> Union[Float[Array, "N N"], LinearOperator]:
        """Add covariance operator to another covariance operator.

        Args:
            other (Union[Float[Array, "N N"], LinearOperator]): Covariance operator to add.

        Returns
        -------
            Union[Float[Array, "N N"], LinearOperator]: Sum of the covariance operators.
        """
        check_shapes_match(self.shape, other.shape)
        return to_linear_operator(other)

    def _add_diagonal(self, other: DiagonalLinearOperator) -> DiagonalLinearOperator:
        """Add diagonal to the covariance operator,  useful for computing, Kxx + Io².

        Args:
            other (DiagonalLinearOperator): Diagonal covariance operator to add to the covariance operator.

        Returns
        -------
            DiagonalLinearOperator: Covariance operator with the diagonal added.
        """
        check_shapes_match(self.shape, other.shape)
        return other

    def __mul__(self, other: ScalarFloat) -> "ZeroLinearOperator":
        """Multiply covariance operator by scalar.

        Args:
            other (ConstantDiagonalLinearOperator): Scalar.

        Returns
        -------
            ZeroLinearOperator: Covariance operator multiplied by a scalar.
        """
        # TODO: check shapes.
        return self

    def __matmul__(
        self, other: Union[LinearOperator, Float[Array, "N M"]]
    ) -> "ZeroLinearOperator":
        """Matrix multiplication.

        Args:
            other (Union[LinearOperator, Float[Array, "N M"]]): Matrix to multiply with.

        Returns
        -------
            Float[Array, "N M"]: Result of matrix multiplication.
        """
        check_shapes_match(self.shape, other.shape)
        return self

    def to_dense(self) -> Float[Array, "N N"]:
        """Construct dense Covariance matrix from the covariance operator.

        Returns
        -------
            Float[Array, "N N"]: Dense covariance matrix.
        """
        return jnp.zeros(self.shape)

    def to_root(self) -> "ZeroLinearOperator":
        """
        Root of the covariance operator.

        Returns
        -------
            ZeroLinearOperator: Root of the covariance operator.
        """
        return self

    def log_det(self) -> ScalarFloat:
        """Log determinant.

        Returns
        -------
            ScalarFloat: Log determinant of the covariance matrix.
        """
        return jnp.log(jnp.array(0.0))

    def inverse(self) -> None:
        """Inverse of the covariance operator.

        Raises
        ------
            RuntimeError: ZeroLinearOperator is not invertible.
        """
        raise RuntimeError("ZeroLinearOperator is not invertible.")

    def solve(self, rhs: Float[Array, "... M"]) -> None:
        """Solve linear system.

        Raises
        ------
            RuntimeError: ZeroLinearOperator is not invertible.
        """
        raise RuntimeError("ZeroLinearOperator is not invertible.")

    @classmethod
    def from_root(cls, root: "ZeroLinearOperator") -> "ZeroLinearOperator":
        """Construct covariance operator from the root.

        Args:
            root (ZeroLinearOperator): Root of the covariance operator.

        Returns
        -------
            ZeroLinearOperator: Covariance operator.
        """
        return root

    @classmethod
    def from_dense(cls, dense: Float[Array, "N N"]) -> "ZeroLinearOperator":
        """Construct covariance operator from the dense matrix.

        Args:
            dense (Float[Array, "N N"]): Dense matrix.

        Returns
        -------
            ZeroLinearOperator: Covariance operator.
        """
        # TODO: check shapes.
        n = dense.shape[0]
        return ZeroLinearOperator(shape=(n, n))


__all__ = [
    "ZeroLinearOperator",
]



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from jax import vmap
from jaxtyping import Float

from gpjax.kernels.computations.base import AbstractKernelComputation
from gpjax.linops import DiagonalLinearOperator
from gpjax.typing import Array


class DiagonalKernelComputation(AbstractKernelComputation):
    r"""Diagonal kernel computation class. Operations with the kernel assume
    a diagonal Gram matrix.
    """

    def gram(self, x: Float[Array, "N D"]) -> DiagonalLinearOperator:
        r"""Compute the Gram matrix.

        For a kernel with diagonal structure, compute the $`N\times N`$ Gram matrix on
        an input matrix of shape $`N\times D`$.

        Args:
            x (Float[Array, "N D"]): The input matrix.

        Returns
        -------
            DiagonalLinearOperator: The computed square Gram matrix.
        """
        return DiagonalLinearOperator(diag=vmap(lambda x: self.kernel(x, x))(x))

    def cross_covariance(
        self, x: Float[Array, "N D"], y: Float[Array, "M D"]
    ) -> Float[Array, "N M"]:
        r"""Compute the cross-covariance matrix.

        For a given kernel, compute the $`N\times M`$ covariance matrix on a pair of
        input matrices of shape $`N\times D`$ and $`M\times D`$.

        Args:
            x (Float[Array,"N D"]): The input matrix.
            y (Float[Array,"M D"]): The input matrix.

        Returns
        -------
            Float[Array, "N M"]: The computed cross-covariance.
        """
        # TODO: This is currently a dense implementation. We should implement a sparse LinearOperator for non-square cross-covariance matrices.
        cross_cov = vmap(lambda x: vmap(lambda y: self.kernel(x, y))(y))(x)
        return cross_cov



# Copyright 2023 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from beartype.typing import (
    Any,
    Callable,
    Union,
)
from jax import lax
from jax.experimental import host_callback
from jaxtyping import Float  # noqa: TCH002
from tqdm.auto import tqdm

from gpjax.typing import Array  # noqa: TCH001


def progress_bar(num_iters: int, log_rate: int) -> Callable:
    r"""Progress bar decorator for the body function of a `jax.lax.scan`.

    Example:
    ```python
        >>> import jax.numpy as jnp
        >>> import jax
        >>>
        >>> carry = jnp.array(0.0)
        >>> iteration_numbers = jnp.arange(100)

        >>> @progress_bar(num_iters=iteration_numbers.shape[0], log_rate=10)
        >>> def body_func(carry, x):
        ...    return carry, x
        >>>
        >>> carry, _ = jax.lax.scan(body_func, carry, iteration_numbers)
    ```

    Adapted from this [excellent blog post](https://www.jeremiecoullon.com/2021/01/29/jax_progress_bar/).

    Might be nice in future to directly create a general purpose `verbose scan`
    inplace of a for a jax.lax.scan, that takes the same arguments as a jax.lax.scan,
    but prints a progress bar.
    """
    tqdm_bars = {}
    remainder = num_iters % log_rate

    """Define a tqdm progress bar."""
    tqdm_bars[0] = tqdm(range(num_iters))
    tqdm_bars[0].set_description("Compiling...", refresh=True)

    def _update_tqdm(args: Any, transform: Any) -> None:
        r"""Update the tqdm progress bar with the latest objective value."""
        value, iter_num = args
        tqdm_bars[0].set_description("Running", refresh=False)
        tqdm_bars[0].update(iter_num)
        tqdm_bars[0].set_postfix({"Value": f"{value: .2f}"})

    def _close_tqdm(args: Any, transform: Any) -> None:
        r"""Close the tqdm progress bar."""
        tqdm_bars[0].close()

    def _callback(cond: bool, func: Callable, arg: Any) -> None:
        r"""Callback a function for a given argument if a condition is true."""
        dummy_result = 0

        def _do_callback(_) -> int:
            """Perform the callback."""
            return host_callback.id_tap(func, arg, result=dummy_result)

        def _not_callback(_) -> int:
            """Do nothing."""
            return dummy_result

        _ = lax.cond(cond, _do_callback, _not_callback, operand=None)

    def _update_progress_bar(value: Float[Array, "1"], iter_num: int) -> None:
        r"""Update the tqdm progress bar."""
        # Conditions for iteration number
        is_multiple: bool = (iter_num % log_rate == 0) & (
            iter_num != num_iters - remainder
        )
        is_remainder: bool = iter_num == num_iters - remainder
        is_last: bool = iter_num == num_iters - 1

        # Update progress bar, if multiple of log_rate
        _callback(is_multiple, _update_tqdm, (value, log_rate))

        # Update progress bar, if remainder
        _callback(is_remainder, _update_tqdm, (value, remainder))

        # Close progress bar, if last iteration
        _callback(is_last, _close_tqdm, None)

    def _progress_bar(body_fun: Callable) -> Callable:
        r"""Decorator that adds a progress bar to `body_fun` used in `jax.lax.scan`."""

        def wrapper_progress_bar(carry: Any, x: Union[tuple, int]) -> Any:
            # Get iteration number
            if type(x) is tuple:
                iter_num, *_ = x
            else:
                iter_num = x

            # Compute iteration step
            result = body_fun(carry, x)

            # Get value
            *_, value = result

            # Update progress bar
            _update_progress_bar(value, iter_num)

            return result

        return wrapper_progress_bar

    return _progress_bar


__all__ = [
    "progress_bar",
]



from abc import abstractmethod
from dataclasses import dataclass

from jax import vmap
import jax.numpy as jnp
import jax.scipy as jsp
import jax.tree_util as jtu
from jaxtyping import Float
import tensorflow_probability.substrates.jax as tfp

from gpjax.base import (
    Module,
    static_field,
)
from gpjax.dataset import Dataset
from gpjax.gaussian_distribution import GaussianDistribution
from gpjax.linops import identity
from gpjax.typing import (
    Array,
    ScalarFloat,
)

tfd = tfp.distributions


@dataclass
class AbstractObjective(Module):
    r"""Abstract base class for objectives."""

    negative: bool = static_field(False)
    constant: float = static_field(init=False, repr=False)

    def __post_init__(self) -> None:
        self.constant = jnp.array(-1.0) if self.negative else jnp.array(1.0)

    def __hash__(self):
        return hash(tuple(jtu.tree_leaves(self)))  # Probably put this on the Module!

    def __call__(self, *args, **kwargs) -> ScalarFloat:
        return self.step(*args, **kwargs)

    @abstractmethod
    def step(self, *args, **kwargs) -> ScalarFloat:
        raise NotImplementedError


class ConjugateMLL(AbstractObjective):
    def step(
        self,
        posterior: "gpjax.gps.ConjugatePosterior",  # noqa: F821
        train_data: Dataset,  # noqa: F821
    ) -> ScalarFloat:
        r"""Evaluate the marginal log-likelihood of the Gaussian process.

        Compute the marginal log-likelihood function of the Gaussian process.
        The returned function can then be used for gradient based optimisation
        of the model's parameters or for model comparison. The implementation
        given here enables exact estimation of the Gaussian process' latent
        function values.

        For a training dataset $`\{x_n, y_n\}_{n=1}^N`$, set of test inputs
        $`\mathbf{x}^{\star}`$ the corresponding latent function evaluations are given
        by $`\mathbf{f}=f(\mathbf{x})`$ and $`\mathbf{f}^{\star}f(\mathbf{x}^{\star})`$,
        the marginal log-likelihood is given by:
        ```math
        \begin{align}
            \log p(\mathbf{y}) & = \int p(\mathbf{y}\mid\mathbf{f})p(\mathbf{f}, \mathbf{f}^{\star}\mathrm{d}\mathbf{f}^{\star}\\
            &=0.5\left(-\mathbf{y}^{\top}\left(k(\mathbf{x}, \mathbf{x}') +\sigma^2\mathbf{I}_N  \right)^{-1}\mathbf{y}-\log\lvert k(\mathbf{x}, \mathbf{x}') + \sigma^2\mathbf{I}_N\rvert - n\log 2\pi \right).
        \end{align}
        ```

        For a given ``ConjugatePosterior`` object, the following code snippet shows
        how the marginal log-likelihood can be evaluated.

        Example:
        ```python
            >>> import gpjax as gpx
            >>>
            >>> xtrain = jnp.linspace(0, 1).reshape(-1, 1)
            >>> ytrain = jnp.sin(xtrain)
            >>> D = gpx.Dataset(X=xtrain, y=ytrain)
            >>>
            >>> meanf = gpx.mean_functions.Constant()
            >>> kernel = gpx.kernels.RBF()
            >>> likelihood = gpx.likelihoods.Gaussian(num_datapoints=D.n)
            >>> prior = gpx.Prior(mean_function = meanf, kernel=kernel)
            >>> posterior = prior * likelihood
            >>>
            >>> mll = gpx.ConjugateMLL(negative=True)
            >>> mll(posterior, train_data = D)
        ```

        Our goal is to maximise the marginal log-likelihood. Therefore, when optimising
        the model's parameters with respect to the parameters, we use the negative
        marginal log-likelihood. This can be realised through

        ```python
            mll = gpx.ConjugateMLL(negative=True)
        ```

        For optimal performance, the marginal log-likelihood should be ``jax.jit``
        compiled.
        ```python
            mll = jit(gpx.ConjugateMLL(negative=True))
        ```

        Args:
            posterior (ConjugatePosterior): The posterior distribution for which
                we want to compute the marginal log-likelihood.
            train_data (Dataset): The training dataset used to compute the
                marginal log-likelihood.

        Returns
        -------
            ScalarFloat: The marginal log-likelihood of the Gaussian process for the
                current parameter set.
        """
        x, y, n = train_data.X, train_data.y, train_data.n

        # Observation noise o²
        obs_noise = posterior.likelihood.obs_noise
        mx = posterior.prior.mean_function(x)

        # Σ = (Kxx + Io²) = LLᵀ
        Kxx = posterior.prior.kernel.gram(x)
        Kxx += identity(n) * posterior.prior.jitter
        Sigma = Kxx + identity(n) * obs_noise

        # p(y | x, θ), where θ are the model hyperparameters:
        mll = GaussianDistribution(jnp.atleast_1d(mx.squeeze()), Sigma)

        return self.constant * (mll.log_prob(jnp.atleast_1d(y.squeeze())).squeeze())


class LogPosteriorDensity(AbstractObjective):
    r"""The log-posterior density of a non-conjugate Gaussian process. This is
    sometimes referred to as the marginal log-likelihood.
    """

    def step(
        self, posterior: "gpjax.gps.NonConjugatePosterior", data: Dataset  # noqa: F821
    ) -> ScalarFloat:
        r"""Evaluate the log-posterior density of a Gaussian process.

        Compute the marginal log-likelihood, or log-posterior density of the Gaussian
        process. The returned function can then be used for gradient based optimisation
        of the model's parameters or for model comparison. The implementation given
        here is general and will work for any likelihood support by GPJax.

        Unlike the marginal_log_likelihood function of the `ConjugatePosterior` object,
        the marginal_log_likelihood function of the `NonConjugatePosterior` object does
        not provide an exact marginal log-likelihood function. Instead, the
        `NonConjugatePosterior` object represents the posterior distributions as a
        function of the model's hyperparameters and the latent function. Markov chain
        Monte Carlo, variational inference, or Laplace approximations can then be used
        to sample from, or optimise an approximation to, the posterior distribution.

        Args:
            posterior (NonConjugatePosterior): The posterior distribution for which
                we want to compute the marginal log-likelihood.
            data (Dataset): The training dataset used to compute the
                marginal log-likelihood.

        Returns
        -------
            ScalarFloat: The log-posterior density of the Gaussian process for the
                current parameter set.
        """
        # Unpack the training data
        x, y, n = data.X, data.y, data.n
        Kxx = posterior.prior.kernel.gram(x)
        Kxx += identity(n) * posterior.prior.jitter
        Lx = Kxx.to_root()

        # Compute the prior mean function
        mx = posterior.prior.mean_function(x)

        # Whitened function values, wx, corresponding to the inputs, x
        wx = posterior.latent

        # f(x) = mx  +  Lx wx
        fx = mx + Lx @ wx

        # p(y | f(x), θ), where θ are the model hyperparameters
        likelihood = posterior.likelihood.link_function(fx)

        # Whitened latent function values prior, p(wx | θ) = N(0, I)
        latent_prior = tfd.Normal(loc=0.0, scale=1.0)

        return self.constant * (
            likelihood.log_prob(y).sum() + latent_prior.log_prob(wx).sum()
        )


NonConjugateMLL = LogPosteriorDensity


class ELBO(AbstractObjective):
    def step(
        self,
        variational_family: "gpjax.variational_families.AbstractVariationalFamily",  # noqa: F821
        train_data: Dataset,
    ) -> ScalarFloat:
        r"""Compute the evidence lower bound of a variational approximation.

        Compute the evidence lower bound under this model. In short, this requires
        evaluating the expectation of the model's log-likelihood under the variational
        approximation. To this, we sum the KL divergence from the variational posterior
        to the prior. When batching occurs, the result is scaled by the batch size
        relative to the full dataset size.

        Args:
            variational_family (AbstractVariationalFamily): The variational
                approximation for whose parameters we should maximise the ELBO with
                respect to.
            train_data (Dataset): The training data for which we should maximise the
                ELBO with respect to.

        Returns
        -------
            ScalarFloat: The evidence lower bound of the variational approximation for
                the current model parameter set.
        """
        # KL[q(f(·)) || p(f(·))]
        kl = variational_family.prior_kl()

        # ∫[log(p(y|f(·))) q(f(·))] df(·)
        var_exp = variational_expectation(variational_family, train_data)

        # For batch size b, we compute  n/b * Σᵢ[ ∫log(p(y|f(xᵢ))) q(f(xᵢ)) df(xᵢ)] - KL[q(f(·)) || p(f(·))]
        return self.constant * (
            jnp.sum(var_exp)
            * variational_family.posterior.likelihood.num_datapoints
            / train_data.n
            - kl
        )


def variational_expectation(
    variational_family: "gpjax.variational_families.AbstractVariationalFamily",  # noqa: F821
    train_data: Dataset,
) -> Float[Array, " N"]:
    r"""Compute the variational expectation.

    Compute the expectation of our model's log-likelihood under our variational
    distribution. Batching can be done here to speed up computation.

    Args:
        variational_family (AbstractVariationalFamily): The variational family that we
            are using to approximate the posterior.
        train_data (Dataset): The batch for which the expectation should be computed
            for.

    Returns
    -------
        Array: The expectation of the model's log-likelihood under our variational
            distribution.
    """
    # Unpack training batch
    x, y = train_data.X, train_data.y

    # Variational distribution q(f(·)) = N(f(·); μ(·), Σ(·, ·))
    q = variational_family

    # Compute variational mean, μ(x), and variance, √diag(Σ(x, x)), at the training
    # inputs, x
    def q_moments(x):
        qx = q(x)
        return qx.mean(), qx.variance()

    mean, variance = vmap(q_moments)(x[:, None])

    # ≈ ∫[log(p(y|f(x))) q(f(x))] df(x)
    expectation = q.posterior.likelihood.expected_log_likelihood(y, mean, variance)
    return expectation


class CollapsedELBO(AbstractObjective):
    r"""The collapsed evidence lower bound.

    Collapsed variational inference for a sparse Gaussian process regression model.
    The key reference is Titsias, (2009) - Variational Learning of Inducing Variables
    in Sparse Gaussian Processes.
    """

    def step(
        self,
        variational_family: "gpjax.variational_families.AbstractVariationalFamily",  # noqa: F821
        train_data: Dataset,
    ) -> ScalarFloat:
        r"""Compute a single step of the collapsed evidence lower bound.

        Compute the evidence lower bound under this model. In short, this requires
        evaluating the expectation of the model's log-likelihood under the variational
        approximation. To this, we sum the KL divergence from the variational posterior
        to the prior. When batching occurs, the result is scaled by the batch size
        relative to the full dataset size.

        Args:
            variational_family (AbstractVariationalFamily): The variational
                approximation for whose parameters we should maximise the ELBO with
                respect to.
            train_data (Dataset): The training data for which we should maximise the
                ELBO with respect to.

        Returns
        -------
            ScalarFloat: The evidence lower bound of the variational approximation for
                the current model parameter set.
        """
        # Unpack training data
        x, y, n = train_data.X, train_data.y, train_data.n

        # Unpack mean function and kernel
        mean_function = variational_family.posterior.prior.mean_function
        kernel = variational_family.posterior.prior.kernel

        m = variational_family.num_inducing

        noise = variational_family.posterior.likelihood.obs_noise
        z = variational_family.inducing_inputs
        Kzz = kernel.gram(z)
        Kzz += identity(m) * variational_family.jitter
        Kzx = kernel.cross_covariance(z, x)
        Kxx_diag = vmap(kernel, in_axes=(0, 0))(x, x)
        μx = mean_function(x)

        Lz = Kzz.to_root()

        # Notation and derivation:
        #
        # Let Q = KxzKzz⁻¹Kzx, we must compute the log normal pdf:
        #
        #   log N(y; μx, o²I + Q) = -nπ - n/2 log|o²I + Q|
        #   - 1/2 (y - μx)ᵀ (o²I + Q)⁻¹ (y - μx).
        #
        # The log determinant |o²I + Q| is computed via applying the matrix determinant
        #   lemma
        #
        #   |o²I + Q| = log|o²I| + log|I + Lz⁻¹ Kzx (o²I)⁻¹ Kxz Lz⁻¹| = log(o²) +  log|B|,
        #
        #   with B = I + AAᵀ and A = Lz⁻¹ Kzx / o.
        #
        # Similarly we apply matrix inversion lemma to invert o²I + Q
        #
        #   (o²I + Q)⁻¹ = (Io²)⁻¹ - (Io²)⁻¹ Kxz Lz⁻ᵀ (I + Lz⁻¹ Kzx (Io²)⁻¹ Kxz Lz⁻ᵀ )⁻¹ Lz⁻¹ Kzx (Io²)⁻¹
        #               = (Io²)⁻¹ - (Io²)⁻¹ oAᵀ (I + oA (Io²)⁻¹ oAᵀ)⁻¹ oA (Io²)⁻¹
        #               = I/o² - Aᵀ B⁻¹ A/o²,
        #
        # giving the quadratic term as
        #
        #   (y - μx)ᵀ (o²I + Q)⁻¹ (y - μx) = [(y - μx)ᵀ(y - µx)  - (y - μx)ᵀ Aᵀ B⁻¹ A (y - μx)]/o²,
        #
        #   with A and B defined as above.

        A = Lz.solve(Kzx) / jnp.sqrt(noise)

        # AAᵀ
        AAT = jnp.matmul(A, A.T)

        # B = I + AAᵀ
        B = jnp.eye(m) + AAT

        # LLᵀ = I + AAᵀ
        L = jnp.linalg.cholesky(B)

        # log|B| = 2 trace(log|L|) = 2 Σᵢ log Lᵢᵢ  [since |B| = |LLᵀ| = |L|²  => log|B| = 2 log|L|, and |L| = Πᵢ Lᵢᵢ]
        log_det_B = 2.0 * jnp.sum(jnp.log(jnp.diagonal(L)))

        diff = y - μx

        # L⁻¹ A (y - μx)
        L_inv_A_diff = jsp.linalg.solve_triangular(L, jnp.matmul(A, diff), lower=True)

        # (y - μx)ᵀ (Io² + Q)⁻¹ (y - μx)
        quad = (jnp.sum(diff**2) - jnp.sum(L_inv_A_diff**2)) / noise

        # 2 * log N(y; μx, Io² + Q)
        two_log_prob = -n * jnp.log(2.0 * jnp.pi * noise) - log_det_B - quad

        # 1/o² tr(Kxx - Q) [Trace law tr(AB) = tr(BA) => tr(KxzKzz⁻¹Kzx) = tr(KxzLz⁻ᵀLz⁻¹Kzx) = tr(Lz⁻¹Kzx KxzLz⁻ᵀ) = trace(o²AAᵀ)]
        two_trace = jnp.sum(Kxx_diag) / noise - jnp.trace(AAT)

        # log N(y; μx, Io² + KxzKzz⁻¹Kzx) - 1/2o² tr(Kxx - KxzKzz⁻¹Kzx)
        return self.constant * (two_log_prob - two_trace).squeeze() / 2.0



# Copyright 2023 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from beartype.typing import (
    Any,
    Callable,
    Optional,
    Tuple,
    TypeVar,
    Union,
)
import jax
from jax._src.random import _check_prng_key
import jax.random as jr
import optax as ox

from gpjax.base import Module
from gpjax.dataset import Dataset
from gpjax.objectives import AbstractObjective
from gpjax.scan import vscan
from gpjax.typing import (
    Array,
    KeyArray,
    ScalarFloat,
)

ModuleModel = TypeVar("ModuleModel", bound=Module)


def fit(  # noqa: PLR0913
    *,
    model: ModuleModel,
    objective: Union[AbstractObjective, Callable[[ModuleModel, Dataset], ScalarFloat]],
    train_data: Dataset,
    optim: ox.GradientTransformation,
    key: KeyArray,
    num_iters: Optional[int] = 100,
    batch_size: Optional[int] = -1,
    log_rate: Optional[int] = 10,
    verbose: Optional[bool] = True,
    unroll: Optional[int] = 1,
    safe: Optional[bool] = True,
) -> Tuple[ModuleModel, Array]:
    r"""Train a Module model with respect to a supplied Objective function.
    Optimisers used here should originate from Optax.

    Example:
    ```python
        >>> import jax.numpy as jnp
        >>> import jax.random as jr
        >>> import optax as ox
        >>> import gpjax as gpx
        >>>
        >>> # (1) Create a dataset:
        >>> X = jnp.linspace(0.0, 10.0, 100)[:, None]
        >>> y = 2.0 * X + 1.0 + 10 * jr.normal(jr.PRNGKey(0), X.shape)
        >>> D = gpx.Dataset(X, y)
        >>>
        >>> # (2) Define your model:
        >>> class LinearModel(gpx.Module):
                weight: float = gpx.param_field()
                bias: float = gpx.param_field()

                def __call__(self, x):
                    return self.weight * x + self.bias

        >>> model = LinearModel(weight=1.0, bias=1.0)
        >>>
        >>> # (3) Define your loss function:
        >>> class MeanSquareError(gpx.AbstractObjective):
                def evaluate(self, model: LinearModel, train_data: gpx.Dataset) -> float:
                    return jnp.mean((train_data.y - model(train_data.X)) ** 2)
        >>>
        >>> loss = MeanSqaureError()
        >>>
        >>> # (4) Train!
        >>> trained_model, history = gpx.fit(
                model=model, objective=loss, train_data=D, optim=ox.sgd(0.001), num_iters=1000
            )
    ```

    Args:
        model (Module): The model Module to be optimised.
        objective (Objective): The objective function that we are optimising with
            respect to.
        train_data (Dataset): The training data to be used for the optimisation.
        optim (GradientTransformation): The Optax optimiser that is to be used for
            learning a parameter set.
        num_iters (Optional[int]): The number of optimisation steps to run. Defaults
            to 100.
        batch_size (Optional[int]): The size of the mini-batch to use. Defaults to -1
            (i.e. full batch).
        key (Optional[KeyArray]): The random key to use for the optimisation batch
            selection. Defaults to jr.PRNGKey(42).
        log_rate (Optional[int]): How frequently the objective function's value should
            be printed. Defaults to 10.
        verbose (Optional[bool]): Whether to print the training loading bar. Defaults
            to True.
        unroll (int): The number of unrolled steps to use for the optimisation.
            Defaults to 1.

    Returns
    -------
        Tuple[Module, Array]: A Tuple comprising the optimised model and training
            history respectively.
    """
    if safe:
        # Check inputs.
        _check_model(model)
        _check_train_data(train_data)
        _check_optim(optim)
        _check_num_iters(num_iters)
        _check_batch_size(batch_size)
        _check_prng_key(key)
        _check_log_rate(log_rate)
        _check_verbose(verbose)

    # Unconstrained space loss function with stop-gradient rule for non-trainable params.
    def loss(model: Module, batch: Dataset) -> ScalarFloat:
        model = model.stop_gradient()
        return objective(model.constrain(), batch)

    # Unconstrained space model.
    model = model.unconstrain()

    # Initialise optimiser state.
    state = optim.init(model)

    # Mini-batch random keys to scan over.
    iter_keys = jr.split(key, num_iters)

    # Optimisation step.
    def step(carry, key):
        model, opt_state = carry

        if batch_size != -1:
            batch = get_batch(train_data, batch_size, key)
        else:
            batch = train_data

        loss_val, loss_gradient = jax.value_and_grad(loss)(model, batch)
        updates, opt_state = optim.update(loss_gradient, opt_state, model)
        model = ox.apply_updates(model, updates)

        carry = model, opt_state
        return carry, loss_val

    # Optimisation scan.
    scan = vscan if verbose else jax.lax.scan

    # Optimisation loop.
    (model, _), history = scan(step, (model, state), (iter_keys), unroll=unroll)

    # Constrained space.
    model = model.constrain()

    return model, history


def get_batch(train_data: Dataset, batch_size: int, key: KeyArray) -> Dataset:
    """Batch the data into mini-batches. Sampling is done with replacement.

    Args:
        train_data (Dataset): The training dataset.
        batch_size (int): The batch size.
        key (KeyArray): The random key to use for the batch selection.

    Returns
    -------
        Dataset: The batched dataset.
    """
    x, y, n = train_data.X, train_data.y, train_data.n

    # Subsample mini-batch indices with replacement.
    indices = jr.choice(key, n, (batch_size,), replace=True)

    return Dataset(X=x[indices], y=y[indices])


def _check_model(model: Any) -> None:
    """Check that the model is of type Module. Check trainables and bijectors tree structure."""
    if not isinstance(model, Module):
        raise TypeError("model must be of type gpjax.Module")


def _check_train_data(train_data: Any) -> None:
    """Check that the train_data is of type Dataset."""
    if not isinstance(train_data, Dataset):
        raise TypeError("train_data must be of type gpjax.Dataset")


def _check_optim(optim: Any) -> None:
    """Check that the optimiser is of type GradientTransformation."""
    if not isinstance(optim, ox.GradientTransformation):
        raise TypeError("optax_optim must be of type optax.GradientTransformation")


def _check_num_iters(num_iters: Any) -> None:
    """Check that the number of iterations is of type int and positive."""
    if not isinstance(num_iters, int):
        raise TypeError("num_iters must be of type int")

    if not num_iters > 0:
        raise ValueError("num_iters must be positive")


def _check_log_rate(log_rate: Any) -> None:
    """Check that the log rate is of type int and positive."""
    if not isinstance(log_rate, int):
        raise TypeError("log_rate must be of type int")

    if not log_rate > 0:
        raise ValueError("log_rate must be positive")


def _check_verbose(verbose: Any) -> None:
    """Check that the verbose is of type bool."""
    if not isinstance(verbose, bool):
        raise TypeError("verbose must be of type bool")


def _check_batch_size(batch_size: Any) -> None:
    """Check that the batch size is of type int and positive if not minus 1."""
    if not isinstance(batch_size, int):
        raise TypeError("batch_size must be of type int")

    if not batch_size == -1 and not batch_size > 0:
        raise ValueError("batch_size must be positive")


__all__ = [
    "fit",
    "get_batch",
]



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from dataclasses import dataclass

import jax.numpy as jnp
from jaxtyping import Float
import tensorflow_probability.substrates.jax.bijectors as tfb

from gpjax.base import (
    param_field,
    static_field,
)
from gpjax.kernels.base import AbstractKernel
from gpjax.kernels.computations import (
    AbstractKernelComputation,
    ConstantDiagonalKernelComputation,
)
from gpjax.typing import (
    Array,
    ScalarFloat,
)


@dataclass
class White(AbstractKernel):
    variance: ScalarFloat = param_field(jnp.array(1.0), bijector=tfb.Softplus())
    compute_engine: AbstractKernelComputation = static_field(
        ConstantDiagonalKernelComputation
    )
    name: str = "White"

    def __call__(self, x: Float[Array, " D"], y: Float[Array, " D"]) -> ScalarFloat:
        r"""Compute the White noise kernel between a pair of arrays.

        Evaluate the kernel on a pair of inputs $`(x, y)`$ with variance $`\sigma^2`$:
        ```math
        k(x, y) = \sigma^2 \delta(x-y)
        ```

        Args:
            x (Float[Array, " D"]): The left hand argument of the kernel function's call.
            y (Float[Array, " D"]): The right hand argument of the kernel function's call.

        Returns
        -------
            ScalarFloat: The value of $`k(x, y)`$.
        """
        K = jnp.all(jnp.equal(x, y)) * self.variance
        return K.squeeze()



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from gpjax.kernels.computations.base import AbstractKernelComputation
from gpjax.kernels.computations.basis_functions import BasisFunctionComputation
from gpjax.kernels.computations.constant_diagonal import (
    ConstantDiagonalKernelComputation,
)
from gpjax.kernels.computations.dense import DenseKernelComputation
from gpjax.kernels.computations.diagonal import DiagonalKernelComputation
from gpjax.kernels.computations.eigen import EigenKernelComputation

__all__ = [
    "AbstractKernelComputation",
    "BasisFunctionComputation",
    "ConstantDiagonalKernelComputation",
    "DenseKernelComputation",
    "DiagonalKernelComputation",
    "EigenKernelComputation",
]



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""JaxKern."""
from gpjax.kernels.approximations import RFF
from gpjax.kernels.base import (
    AbstractKernel,
    ProductKernel,
    SumKernel,
)
from gpjax.kernels.computations import (
    BasisFunctionComputation,
    ConstantDiagonalKernelComputation,
    DenseKernelComputation,
    DiagonalKernelComputation,
    EigenKernelComputation,
)
from gpjax.kernels.non_euclidean import GraphKernel
from gpjax.kernels.nonstationary import (
    ArcCosine,
    Linear,
    Polynomial,
)
from gpjax.kernels.stationary import (
    RBF,
    Matern12,
    Matern32,
    Matern52,
    Periodic,
    PoweredExponential,
    RationalQuadratic,
    White,
)

__all__ = [
    "AbstractKernel",
    "ArcCosine",
    "RBF",
    "GraphKernel",
    "Matern12",
    "Matern32",
    "Matern52",
    "Linear",
    "Polynomial",
    "ProductKernel",
    "SumKernel",
    "DenseKernelComputation",
    "DiagonalKernelComputation",
    "ConstantDiagonalKernelComputation",
    "EigenKernelComputation",
    "PoweredExponential",
    "Periodic",
    "RationalQuadratic",
    "White",
    "BasisFunctionComputation",
    "RFF",
]



# Copyright 2022 The JaxLinOp Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


import abc
from dataclasses import dataclass

from beartype.typing import (
    Any,
    Generic,
    Iterable,
    Mapping,
    Tuple,
    Type,
    TypeVar,
    Union,
)
import jax.numpy as jnp
from jaxtyping import Float
from simple_pytree import (
    Pytree,
    static_field,
)

from gpjax.typing import (
    Array,
    ScalarFloat,
)

# Generic type.
T = TypeVar("T")

# Generic nested type.
NestedT = Union[T, Iterable["NestedT"], Mapping[Any, "NestedT"]]

# Nested types.
DTypes = Union[Type[jnp.float32], Type[jnp.float64], Type[jnp.int32], Type[jnp.int64]]
ShapeT = TypeVar("ShapeT", bound=NestedT[Tuple[int, ...]])
DTypeT = TypeVar("DTypeT", bound=NestedT[DTypes])

# The Generic type is used for type checking the LinearOperator's shape and datatype.
# `static_field` is used to mark nodes of the PyTree that don't change under JAX transformations.
# this is important, so that we e.g., don't take the gradient with respect to the shape!


@dataclass
class LinearOperator(Pytree, Generic[ShapeT, DTypeT]):
    """Linear operator base class."""

    shape: ShapeT = static_field()
    dtype: DTypeT = static_field()

    def __repr__(self) -> str:
        """Linear operator representation."""
        return f"{type(self).__name__}(shape={self.shape}, dtype={self.dtype.__name__})"

    @property
    def ndim(self) -> int:
        """Linear operator dimension."""
        return len(self.shape)

    @property
    def T(self) -> "LinearOperator":
        """Transpose linear operator. Currently, we assume all linear operators are square and symmetric."""
        return self

    def __sub__(
        self, other: Union["LinearOperator", Float[Array, "N N"]]
    ) -> "LinearOperator":
        """Subtract linear operator."""
        return self + (other * -1)

    def __rsub__(
        self, other: Union["LinearOperator", Float[Array, "N N"]]
    ) -> "LinearOperator":
        """Reimplimentation of subtract linear operator."""
        return (self * -1) + other

    def __add__(
        self, other: Union["LinearOperator", Float[Array, "N N"]]
    ) -> "LinearOperator":
        """Add linear operator."""
        raise NotImplementedError

    def __radd__(
        self, other: Union["LinearOperator", Float[Array, "N N"]]
    ) -> "LinearOperator":
        """Reimplimentation of add linear operator."""
        return self + other

    @abc.abstractmethod
    def __mul__(self, other: ScalarFloat) -> "LinearOperator":
        """Multiply linear operator by scalar."""
        raise NotImplementedError

    def __rmul__(self, other: ScalarFloat) -> "LinearOperator":
        """Reimplimentation of multiply linear operator by scalar."""
        return self.__mul__(other)

    @abc.abstractmethod
    def _add_diagonal(
        self,
        other: "gpjax.linops.diagonal_linear_operator.DiagonalLinearOperator",  # noqa: F821
    ) -> "LinearOperator":
        """Add diagonal linear operator to a linear operator, useful e.g., for adding jitter."""
        return NotImplementedError

    @abc.abstractmethod
    def __matmul__(
        self, other: Union["LinearOperator", Float[Array, "N M"]]
    ) -> Union["LinearOperator", Float[Array, "N M"]]:
        """Matrix multiplication."""
        raise NotImplementedError

    def __rmatmul__(
        self, other: Union["LinearOperator", Float[Array, "N M"]]
    ) -> Union["LinearOperator", Float[Array, "N M"]]:
        """Reimplimentation of matrix multiplication."""
        # Exploit the fact that linear operators are square and symmetric.
        if other.ndim == 1:
            return self.T @ other
        return (self.T @ other.T).T

    @abc.abstractmethod
    def diagonal(self) -> Float[Array, " N"]:
        """Diagonal of the linear operator.

        Returns
        -------
            Float[Array, " N"]: Diagonal of the linear operator.
        """
        raise NotImplementedError

    def trace(self) -> ScalarFloat:
        """Trace of the linear matrix.

        Returns
        -------
            ScalarFloat: Trace of the linear matrix.
        """
        return jnp.sum(self.diagonal())

    def log_det(self) -> ScalarFloat:
        """Log determinant of the linear matrix. Default implementation uses dense Cholesky decomposition.

        Returns
        -------
            ScalarFloat: Log determinant of the linear matrix.
        """
        root = self.to_root()

        return 2.0 * jnp.sum(jnp.log(root.diagonal()))

    def to_root(self) -> "LinearOperator":
        """Compute the root of the linear operator via the Cholesky decomposition.

        Returns
        -------
            Float[Array, "N N"]: Lower Cholesky decomposition of the linear operator.
        """
        from gpjax.linops.triangular_linear_operator import (
            LowerTriangularLinearOperator,
        )

        L = jnp.linalg.cholesky(self.to_dense())

        return LowerTriangularLinearOperator.from_dense(L)

    def inverse(self) -> "LinearOperator":
        """Inverse of the linear matrix. Default implementation uses dense Cholesky decomposition.

        Returns
        -------
            LinearOperator: Inverse of the linear matrix.
        """
        from gpjax.linops.dense_linear_operator import DenseLinearOperator

        n = self.shape[0]

        return DenseLinearOperator(self.solve(jnp.eye(n)))

    def solve(self, rhs: Float[Array, "... M"]) -> Float[Array, "... M"]:
        """Solve linear system. Default implementation uses dense Cholesky decomposition.

        Args:
            rhs (Float[Array, "N M"]): Right hand side of the linear system.

        Returns
        -------
            Float[Array, "N M"]: Solution of the linear system.
        """
        root = self.to_root()
        rootT = root.T

        return rootT.solve(root.solve(rhs))

    @abc.abstractmethod
    def to_dense(self) -> Float[Array, "N N"]:
        """Construct dense matrix from the linear operator.

        Returns
        -------
            Float[Array, "N N"]: Dense linear matrix.
        """
        raise NotImplementedError

    @classmethod
    def from_dense(cls, dense: Float[Array, "N N"]) -> "LinearOperator":
        """Construct linear operator from dense matrix.

        Args:
            dense (Float[Array, "N N"]): Dense matrix.

        Returns
        -------
            LinearOperator: Linear operator.
        """
        raise NotImplementedError


__all__ = [
    "LinearOperator",
]



# Copyright 2022 The JaxLinOp Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from dataclasses import dataclass

from beartype.typing import (
    Any,
    Union,
)
import jax.numpy as jnp
from jaxtyping import Float

from gpjax.base import static_field
from gpjax.linops.diagonal_linear_operator import DiagonalLinearOperator
from gpjax.linops.linear_operator import LinearOperator
from gpjax.typing import (
    Array,
    ScalarFloat,
)


def _check_args(value: Any, size: Any) -> None:
    if not isinstance(size, int):
        raise ValueError(f"`length` must be an integer, but `length = {size}`.")

    if value.ndim != 1:
        raise ValueError(
            "`value` must be one dimensional scalar, but `value.shape ="
            f" {value.shape}`."
        )


@dataclass
class ConstantDiagonalLinearOperator(DiagonalLinearOperator):
    value: Float[Array, "1"]
    size: int = static_field()

    def __init__(
        self, value: Float[Array, "1"], size: int, dtype: jnp.dtype = None
    ) -> None:
        """Initialize the constant diagonal linear operator.

        Args:
            value (Float[Array, "1"]): Constant value of the diagonal.
            size (int): Size of the diagonal.
        """
        _check_args(value, size)

        if dtype is not None:
            value = value.astype(dtype)

        self.value = value
        self.size = size
        self.shape = (size, size)
        self.dtype = value.dtype

    def __add__(
        self, other: Union[Float[Array, "N N"], LinearOperator]
    ) -> LinearOperator:
        if isinstance(other, ConstantDiagonalLinearOperator):
            if other.size == self.size:
                return ConstantDiagonalLinearOperator(
                    value=self.value + other.value, size=self.size
                )

            raise ValueError(
                f"`length` must be the same, but `length = {self.size}` and `length ="
                f" {other.size}`."
            )

        else:
            return super().__add__(other)

    def __mul__(self, other: Union[ScalarFloat, Float[Array, "1"]]) -> LinearOperator:
        """Multiply covariance operator by scalar.

        Args:
            other (LinearOperator): Scalar.

        Returns
        -------
            LinearOperator: Covariance operator multiplied by a scalar.
        """
        return ConstantDiagonalLinearOperator(value=self.value * other, size=self.size)

    def _add_diagonal(self, other: DiagonalLinearOperator) -> LinearOperator:
        """Add diagonal to the covariance operator,  useful for computing, Kxx + Io².

        Args:
            other (DiagonalLinearOperator): Diagonal covariance operator to add to the covariance operator.

        Returns
        -------
            LinearOperator: Covariance operator with the diagonal added.
        """
        if isinstance(other, ConstantDiagonalLinearOperator):
            if other.size == self.size:
                return ConstantDiagonalLinearOperator(
                    value=self.value + other.value, size=self.size
                )

            raise ValueError(
                f"`length` must be the same, but `length = {self.size}` and `length ="
                f" {other.size}`."
            )

        else:
            return super()._add_diagonal(other)

    def diagonal(self) -> Float[Array, " N"]:
        """Diagonal of the covariance operator."""
        return self.value * jnp.ones(self.size)

    def to_root(self) -> "ConstantDiagonalLinearOperator":
        """
        Lower triangular.

        Returns
        -------
            Float[Array, "N N"]: Lower triangular matrix.
        """
        return ConstantDiagonalLinearOperator(
            value=jnp.sqrt(self.value), size=self.size
        )

    def log_det(self) -> ScalarFloat:
        """Log determinant.

        Returns
        -------
            ScalarFloat: Log determinant of the covariance matrix.
        """
        return 2.0 * self.size * jnp.log(self.value.squeeze())

    def inverse(self) -> "ConstantDiagonalLinearOperator":
        """Inverse of the covariance operator.

        Returns
        -------
            DiagonalLinearOperator: Inverse of the covariance operator.
        """
        return ConstantDiagonalLinearOperator(value=1.0 / self.value, size=self.size)

    def solve(self, rhs: Float[Array, "... M"]) -> Float[Array, "... M"]:
        """Solve linear system.

        Args:
            rhs (Float[Array, "N M"]): Right hand side of the linear system.

        Returns
        -------
            Float[Array, "N M"]: Solution of the linear system.
        """
        return rhs / self.value

    @classmethod
    def from_dense(cls, dense: Float[Array, "N N"]) -> "ConstantDiagonalLinearOperator":
        """Construct covariance operator from dense matrix.

        Args:
            dense (Float[Array, "N N"]): Dense matrix.

        Returns
        -------
            DiagonalLinearOperator: Covariance operator.
        """
        return ConstantDiagonalLinearOperator(
            value=jnp.atleast_1d(dense[0, 0]), size=dense.shape[0]
        )

    @classmethod
    def from_root(
        cls, root: "ConstantDiagonalLinearOperator"
    ) -> "ConstantDiagonalLinearOperator":
        """Construct covariance operator from root.

        Args:
            root (ConstantDiagonalLinearOperator): Root of the covariance operator.

        Returns
        -------
            ConstantDiagonalLinearOperator: Covariance operator.
        """
        return ConstantDiagonalLinearOperator(value=root.value**2, size=root.size)


__all__ = [
    "ConstantDiagonalLinearOperator",
]



from abc import abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING

from beartype.typing import Callable
import jax.numpy as jnp
from jaxtyping import Float
import numpy as np

from gpjax.typing import Array

if TYPE_CHECKING:
    import gpjax.likelihoods


@dataclass
class AbstractIntegrator:
    r"""Base class for integrators."""

    @abstractmethod
    def integrate(
        self,
        fun: Callable,
        y: Float[Array, "N D"],
        mean: Float[Array, "N D"],
        variance: Float[Array, "N D"],
        likelihood: "gpjax.likelihoods.AbstractLikelihood" = None,
    ) -> Float[Array, " N"]:
        r"""Integrate a function with respect to a Gaussian distribution.

        Typically, the function will be the likelihood function and the mean
        and variance will be the parameters of the variational distribution.

        Args:
            fun (Callable): The function to be integrated.
            y (Float[Array, 'N D']): The observed response variable.
            mean (Float[Array, 'N D']): The mean of the variational distribution.
            variance (Float[Array, 'N D']): The variance of the variational
                distribution.
            likelihood (AbstractLikelihood): The likelihood function.
        """
        raise NotImplementedError("self.integrate not implemented")

    def __call__(
        self,
        fun: Callable,
        y: Float[Array, "N D"],
        mean: Float[Array, "N D"],
        variance: Float[Array, "N D"],
        likelihood: "gpjax.likelihoods.AbstractLikelihood" = None,
    ) -> Float[Array, " N"]:
        r"""Integrate a function with respect to a Gaussian distribution.

        Typically, the function will be the likelihood function and the mean
        and variance will be the parameters of the variational distribution.

        Args:
            fun (Callable): The function to be integrated.
            y (Float[Array, 'N D']): The observed response variable.
            mean (Float[Array, 'N D']): The mean of the variational distribution.
            variance (Float[Array, 'N D']): The variance of the variational
                distribution.
            likelihood (AbstractLikelihood): The likelihood function.
        """
        return self.integrate(fun, y, mean, variance, likelihood)


@dataclass
class GHQuadratureIntegrator(AbstractIntegrator):
    r"""Compute an integral using Gauss-Hermite quadrature.

    Gauss-Hermite quadrature is a method for approximating integrals through a
    weighted sum of function evaluations at specific points
    ```math
    \int F(t)\exp(-t^2)\mathrm{d}t \approx \sum_{j=1}^J w_j F(t_j)
    ```
    where $`t_j`$ and $`w_j`$ are the roots and weights of the $`J`$-th order Hermite
    polynomial $`H_J(t)`$ that we can look up in table
    [link](https://keisan.casio.com/exec/system/1281195844).
    """
    num_points: int = 20

    def integrate(
        self,
        fun: Callable,
        y: Float[Array, "N D"],
        mean: Float[Array, "N D"],
        variance: Float[Array, "N D"],
        likelihood: "gpjax.likelihoods.AbstractLikelihood" = None,
    ) -> Float[Array, " N"]:
        r"""Compute a quadrature integral.

        Args:
            fun (Callable): The likelihood to be integrated.
            y (Float[Array, 'N D']): The observed response variable.
            mean (Float[Array, 'N D']): The mean of the variational distribution.
            variance (Float[Array, 'N D']): The variance of the variational
                distribution.
            likelihood (AbstractLikelihood): The likelihood function.

        Returns:
            Float[Array, 'N']: The expected log likelihood.
        """
        gh_points, gh_weights = np.polynomial.hermite.hermgauss(self.num_points)
        sd = jnp.sqrt(variance)
        X = mean + jnp.sqrt(2.0) * sd * gh_points
        W = gh_weights / jnp.sqrt(jnp.pi)
        val = jnp.sum(fun(X, y) * W, axis=1)
        return val


@dataclass
class AnalyticalGaussianIntegrator(AbstractIntegrator):
    r"""Compute the analytical integral of a Gaussian likelihood.

    When the likelihood function is Gaussian, the integral can be computed in closed
    form. For a Gaussian likelihood $`p(y|f) = \mathcal{N}(y|f, \sigma^2)`$ and a
    variational distribution $`q(f) = \mathcal{N}(f|m, s)`$, the expected
    log-likelihood is given by
    ```math
    \mathbb{E}_{q(f)}[\log p(y|f)] = -\frac{1}{2}\left(\log(2\pi\sigma^2) + \frac{1}{\sigma^2}\mathbb{E}_{q(f)}[(y-f)^2 + s]\right)
    """

    def integrate(
        self,
        fun: Callable,
        y: Float[Array, "N D"],
        mean: Float[Array, "N D"],
        variance: Float[Array, "N D"],
        likelihood: "gpjax.likelihoods.Gaussian" = None,
    ) -> Float[Array, " N"]:
        r"""Compute a Gaussian integral.

        Args:
            fun (Callable): The Gaussian likelihood to be integrated.
            y (Float[Array, 'N D']): The observed response variable.
            mean (Float[Array, 'N D']): The mean of the variational distribution.
            variance (Float[Array, 'N D']): The variance of the variational
                distribution.
            likelihood (Gaussian): The Gaussian likelihood function.

        Returns:
            Float[Array, 'N']: The expected log likelihood.
        """
        obs_noise = likelihood.obs_noise.squeeze()
        sq_error = jnp.square(y - mean)
        log2pi = jnp.log(2.0 * jnp.pi)
        val = jnp.sum(
            log2pi + jnp.log(obs_noise) + (sq_error + variance) / obs_noise, axis=1
        )
        return -0.5 * val


__all__ = [
    "AbstractIntegrator",
    "GHQuadratureIntegrator",
    "AnalyticalGaussianIntegrator",
]



# Copyright 2022 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
from gpjax import integrators
from gpjax.base import (
    Module,
    param_field,
)
from gpjax.citation import cite
from gpjax.dataset import Dataset
from gpjax.fit import fit
from gpjax.gps import (
    Prior,
    construct_posterior,
)
from gpjax.kernels import (
    RBF,
    RFF,
    AbstractKernel,
    BasisFunctionComputation,
    ConstantDiagonalKernelComputation,
    DenseKernelComputation,
    DiagonalKernelComputation,
    EigenKernelComputation,
    GraphKernel,
    Linear,
    Matern12,
    Matern32,
    Matern52,
    Periodic,
    Polynomial,
    PoweredExponential,
    ProductKernel,
    RationalQuadratic,
    SumKernel,
    White,
)
from gpjax.likelihoods import (
    Bernoulli,
    Gaussian,
    Poisson,
)
from gpjax.mean_functions import (
    Constant,
    Zero,
)
from gpjax.objectives import (
    ELBO,
    CollapsedELBO,
    ConjugateMLL,
    LogPosteriorDensity,
    NonConjugateMLL,
)
from gpjax.variational_families import (
    CollapsedVariationalGaussian,
    ExpectationVariationalGaussian,
    NaturalVariationalGaussian,
    VariationalGaussian,
    WhitenedVariationalGaussian,
)

__license__ = "MIT"
__description__ = "Didactic Gaussian processes in JAX"
__url__ = "https://github.com/JaxGaussianProcesses/GPJax"
__contributors__ = "https://github.com/JaxGaussianProcesses/GPJax/graphs/contributors"
__version__ = "0.0.0"

__all__ = [
    "Module",
    "param_field",
    "cite",
    "kernels",
    "fit",
    "Prior",
    "construct_posterior",
    "integrators",
    "RBF",
    "GraphKernel",
    "Matern12",
    "Matern32",
    "Matern52",
    "Polynomial",
    "ProductKernel",
    "SumKernel",
    "Bernoulli",
    "Gaussian",
    "Poisson",
    "Constant",
    "Zero",
    "Dataset",
    "CollapsedVariationalGaussian",
    "ExpectationVariationalGaussian",
    "NaturalVariationalGaussian",
    "VariationalGaussian",
    "WhitenedVariationalGaussian",
    "CollapsedVI",
    "StochasticVI",
    "ConjugateMLL",
    "NonConjugateMLL",
    "LogPosteriorDensity",
    "CollapsedELBO",
    "ELBO",
    "AbstractKernel",
    "Linear",
    "DenseKernelComputation",
    "DiagonalKernelComputation",
    "ConstantDiagonalKernelComputation",
    "EigenKernelComputation",
    "PoweredExponential",
    "Periodic",
    "RationalQuadratic",
    "White",
    "BasisFunctionComputation",
    "RFF",
]



# Copyright 2023 The GPJax Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from beartype.typing import (
    Callable,
    Union,
)
from jax.random import KeyArray as JAXKeyArray
from jaxtyping import (
    Array as JAXArray,
    Bool,
    Float,
    Int,
    UInt32,
)
from numpy import ndarray as NumpyArray

OldKeyArray = UInt32[JAXArray, "2"]
KeyArray = Union[
    OldKeyArray, JAXKeyArray
]  # for compatibility regardless of enable_custom_prng setting

Array = Union[JAXArray, NumpyArray]

ScalarBool = Union[bool, Bool[Array, ""]]
ScalarInt = Union[int, Int[Array, ""]]
ScalarFloat = Union[float, Float[Array, ""]]

VecNOrMatNM = Union[Float[Array, " N"], Float[Array, "N M"]]

FunctionalSample = Callable[[Float[Array, "N D"]], Float[Array, "N B"]]
r""" Type alias for functions representing $`B`$ samples from a model, to be evaluated on
any set of $`N`$ inputs (of dimension $`D`$) and returning the evaluations of each
(potentially approximate) sample draw across these inputs.
"""

__all__ = ["KeyArray", "ScalarBool", "ScalarInt", "ScalarFloat", "FunctionalSample"]



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from dataclasses import dataclass

from beartype.typing import Union
import jax.numpy as jnp
from jaxtyping import Float
import tensorflow_probability.substrates.jax.bijectors as tfb

from gpjax.base import (
    param_field,
    static_field,
)
from gpjax.kernels.base import AbstractKernel
from gpjax.typing import (
    Array,
    ScalarFloat,
    ScalarInt,
)


@dataclass
class ArcCosine(AbstractKernel):
    r"""The ArCosine kernel.

    This kernel is non-stationary and resembles the behavior of neural networks.
    See Section 3.1 of
    [Cho and Saul (2011)](https://arxiv.org/abs/1112.3712) for
    additional details.
    """

    order: ScalarInt = static_field(0)
    variance: ScalarFloat = param_field(jnp.array(1.0), bijector=tfb.Softplus())
    weight_variance: Union[ScalarFloat, Float[Array, " D"]] = param_field(
        jnp.array(1.0), bijector=tfb.Softplus()
    )
    bias_variance: ScalarFloat = param_field(jnp.array(1.0), bijector=tfb.Softplus())

    def __post_init__(self):
        if self.order not in [0, 1, 2]:
            raise ValueError("ArcCosine kernel only implemented for orders 0, 1 and 2.")

        self.name = f"ArcCosine (order {self.order})"

    def __call__(self, x: Float[Array, " D"], y: Float[Array, " D"]) -> ScalarFloat:
        r"""Evaluate the kernel on a pair of inputs $`(x, y)`$

        Args:
            x (Float[Array, "D"]): The left hand argument of the kernel function's
                call.
            y (Float[Array, "D"]): The right hand argument of the kernel function's
                call

        Returns
        -------
            ScalarFloat: The value of $`k(x, y)`$.
        """

        x = self.slice_input(x)
        y = self.slice_input(y)

        x_x = self._weighted_prod(x, x)
        x_y = self._weighted_prod(x, y)
        y_y = self._weighted_prod(y, y)

        cos_theta = x_y / jnp.sqrt(x_x * y_y)
        jitter = 1e-15  # improve numerical stability
        theta = jnp.arccos(jitter + (1 - 2 * jitter) * cos_theta)

        K = self._J(theta)
        K *= jnp.sqrt(x_x) ** self.order
        K *= jnp.sqrt(y_y) ** self.order
        K *= self.variance / jnp.pi

        return K.squeeze()

    def _weighted_prod(
        self, x: Float[Array, " D"], y: Float[Array, " D"]
    ) -> ScalarFloat:
        r"""Calculate the weighted product between two arguments.

        Args:
            x (Float[Array, "D"]): The left hand argument.
            y (Float[Array, "D"]): The right hand argument.
        Returns
        -------
            ScalarFloat: The value of the weighted product between the two arguments``.
        """
        return jnp.inner(self.weight_variance * x, y) + self.bias_variance

    def _J(self, theta: ScalarFloat) -> ScalarFloat:
        r"""Evaluate the angular dependency function corresponding to the desired order.

        Args:
            theta (Float[Array, "1"]): The weighted angle between inputs.

        Returns
        -------
            Float[Array, "1"]: The value of the angular dependency function`.
        """

        if self.order == 0:
            return jnp.pi - theta
        elif self.order == 1:
            return jnp.sin(theta) + (jnp.pi - theta) * jnp.cos(theta)
        else:
            return 3.0 * jnp.sin(theta) * jnp.cos(theta) + (jnp.pi - theta) * (
                1.0 + 2.0 * jnp.cos(theta) ** 2
            )



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from dataclasses import dataclass

from beartype.typing import Optional
import jax.numpy as jnp
from jaxtyping import Num
from simple_pytree import Pytree

from gpjax.typing import Array


@dataclass
class Dataset(Pytree):
    r"""Base class for datasets.

    Attributes
    ----------
        X (Optional[Num[Array, "N D"]]): Input data.
        y (Optional[Num[Array, "N Q"]]): Output data.
    """

    X: Optional[Num[Array, "N D"]] = None
    y: Optional[Num[Array, "N Q"]] = None

    def __post_init__(self) -> None:
        r"""Checks that the shapes of $`X`$ and $`y`$ are compatible."""
        _check_shape(self.X, self.y)

    def __repr__(self) -> str:
        r"""Returns a string representation of the dataset."""
        repr = (
            f"- Number of observations: {self.n}\n- Input dimension:"
            f" {self.in_dim}\n- Output dimension: {self.out_dim}"
        )
        return repr

    def is_supervised(self) -> bool:
        r"""Returns `True` if the dataset is supervised."""
        return self.X is not None and self.y is not None

    def is_unsupervised(self) -> bool:
        r"""Returns `True` if the dataset is unsupervised."""
        return self.X is None and self.y is not None

    def __add__(self, other: "Dataset") -> "Dataset":
        r"""Combine two datasets. Right hand dataset is stacked beneath the left."""
        X = None
        y = None

        if self.X is not None and other.X is not None:
            X = jnp.concatenate((self.X, other.X))

        if self.y is not None and other.y is not None:
            y = jnp.concatenate((self.y, other.y))

        return Dataset(X=X, y=y)

    @property
    def n(self) -> int:
        r"""Number of observations."""
        return self.X.shape[0]

    @property
    def in_dim(self) -> int:
        r"""Dimension of the inputs, $`X`$."""
        return self.X.shape[1]

    @property
    def out_dim(self) -> int:
        r"""Dimension of the outputs, $`y`$."""
        return self.y.shape[1]


def _check_shape(
    X: Optional[Num[Array, "..."]], y: Optional[Num[Array, "..."]]
) -> None:
    r"""Checks that the shapes of $`X`$ and $`y`$ are compatible."""
    if X is not None and y is not None and X.shape[0] != y.shape[0]:
        raise ValueError(
            "Inputs, X, and outputs, y, must have the same number of rows."
            f" Got X.shape={X.shape} and y.shape={y.shape}."
        )

    if X is not None and X.ndim != 2:
        raise ValueError(
            f"Inputs, X, must be a 2-dimensional array. Got X.ndim={X.ndim}."
        )

    if y is not None and y.ndim != 2:
        raise ValueError(
            f"Outputs, y, must be a 2-dimensional array. Got y.ndim={y.ndim}."
        )


__all__ = [
    "Dataset",
]



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


__all__ = ["Module", "meta_leaves", "meta_flatten", "meta_map", "meta", "static_field"]

from copy import (
    copy,
    deepcopy,
)
import dataclasses
import os

from beartype.typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Tuple,
    TypeVar,
    Union,
)
import jax
from jax._src.tree_util import _registry
import jax.tree_util as jtu
from orbax.checkpoint import (
    ArrayRestoreArgs,
    Checkpointer,
    PyTreeCheckpointHandler,
    RestoreArgs,
    SaveArgs,
)
from simple_pytree import Pytree
import tensorflow_probability.substrates.jax.bijectors as tfb

Self = TypeVar("Self")


def static_field(  # noqa: PLR0913
    default: Any = dataclasses.MISSING,
    *,
    default_factory: Any = dataclasses.MISSING,
    init: bool = True,
    repr: bool = True,
    hash: Optional[bool] = None,
    compare: bool = True,
    metadata: Optional[Mapping[str, Any]] = None,
):
    metadata = {} if metadata is None else dict(metadata)

    if "pytree_node" in metadata:
        raise ValueError("Cannot use metadata with `pytree_node` already set.")

    metadata["pytree_node"] = False

    if (
        default is not dataclasses.MISSING
        and default_factory is not dataclasses.MISSING
    ):
        raise ValueError("Cannot specify both default and default_factory.")

    if default is not dataclasses.MISSING:
        default_factory = lambda: default

    return dataclasses.field(
        default_factory=default_factory,
        init=init,
        repr=repr,
        hash=hash,
        compare=compare,
        metadata=metadata,
    )


class Module(Pytree):
    _pytree__meta: Dict[str, Any] = static_field()

    def __init_subclass__(cls, mutable: bool = False):
        cls._pytree__meta = {}
        super().__init_subclass__(mutable=mutable)
        class_vars = vars(cls)
        for field, value in class_vars.items():
            if (
                field not in cls._pytree__static_fields
                and isinstance(value, dataclasses.Field)
                and value.metadata is not None
            ):
                cls._pytree__meta[field] = {**value.metadata}

    def replace(self: Self, **kwargs: Any) -> Self:
        """
        Replace the values of the fields of the object.

        Args:
            **kwargs: keyword arguments to replace the fields of the object.

        Returns
        -------
            Module: with the fields replaced.
        """
        fields = vars(self)
        for key in kwargs:
            if key not in fields:
                raise ValueError(f"'{key}' is not a field of {type(self).__name__}")

        pytree = copy(self)
        pytree.__dict__.update(kwargs)
        return pytree

    def replace_meta(self: Self, **kwargs: Any) -> Self:
        """
        Replace the metadata of the fields.

        Args:
            **kwargs: keyword arguments to replace the metadata of the fields of the object.

        Returns
        -------
            Module: with the metadata of the fields replaced.
        """
        fields = vars(self)
        for key in kwargs:
            if key not in fields:
                raise ValueError(f"'{key}' is not a field of {type(self).__name__}")

        pytree = copy(self)
        pytree.__dict__.update(_pytree__meta={**pytree._pytree__meta, **kwargs})
        return pytree

    def update_meta(self: Self, **kwargs: Any) -> Self:
        """
        Update the metadata of the fields. The metadata must already exist.

        Args:
            **kwargs: keyword arguments to replace the fields of the object.

        Returns
        -------
            Module: with the fields replaced.
        """
        fields = vars(self)
        for key in kwargs:
            if key not in fields:
                raise ValueError(f"'{key}' is not a field of {type(self).__name__}")

        pytree = copy(self)
        new = deepcopy(pytree._pytree__meta)
        for key, value in kwargs.items():
            if key in new:
                new[key].update(value)
            else:
                new[key] = value
        pytree.__dict__.update(_pytree__meta=new)
        return pytree

    def replace_trainable(self: Self, **kwargs: Dict[str, bool]) -> Self:
        """Replace the trainability status of local nodes of the Module."""
        return self.update_meta(**{k: {"trainable": v} for k, v in kwargs.items()})

    def replace_bijector(self: Self, **kwargs: Dict[str, tfb.Bijector]) -> Self:
        """Replace the bijectors of local nodes of the Module."""
        return self.update_meta(**{k: {"bijector": v} for k, v in kwargs.items()})

    def constrain(self: Self) -> Self:
        """Transform model parameters to the constrained space according to their defined bijectors.

        Returns
        -------
            Module: transformed to the constrained space.
        """

        def _apply_constrain(meta_leaf):
            meta, leaf = meta_leaf

            if meta is None:
                return leaf

            return meta.get("bijector", tfb.Identity()).forward(leaf)

        return meta_map(_apply_constrain, self)

    def unconstrain(self: Self) -> Self:
        """Transform model parameters to the unconstrained space according to their defined bijectors.

        Returns
        -------
            Module: transformed to the unconstrained space.
        """

        def _apply_unconstrain(meta_leaf):
            meta, leaf = meta_leaf

            if meta is None:
                return leaf

            return meta.get("bijector", tfb.Identity()).inverse(leaf)

        return meta_map(_apply_unconstrain, self)

    def stop_gradient(self: Self) -> Self:
        """Stop gradients flowing through the Module.

        Returns
        -------
            Module: with gradients stopped.
        """

        # 🛑 Stop gradients flowing through a given leaf if it is not trainable.
        def _stop_grad(leaf: jax.Array, trainable: bool) -> jax.Array:
            return jax.lax.cond(trainable, lambda x: x, jax.lax.stop_gradient, leaf)

        def _apply_stop_grad(meta_leaf):
            meta, leaf = meta_leaf

            if meta is None:
                return leaf

            return _stop_grad(leaf, meta.get("trainable", True))

        return meta_map(_apply_stop_grad, self)

    def trainables(self: Self) -> Self:
        def _get_trainables(meta_leaf):
            meta, leaf = meta_leaf
            if meta is None:
                return True

            return meta.get("trainable", True)

        return meta_map(_get_trainables, self)


def _toplevel_meta(pytree: Any) -> List[Optional[Dict[str, Any]]]:
    """Unpacks a list of meta corresponding to the top-level nodes of the pytree.

    Args:
        pytree (Any): pytree to unpack the meta from.

    Returns
    -------
        List[Dict[str, Any]]: meta of the top-level nodes of the pytree.
    """
    if isinstance(pytree, Iterable):
        return [None] * len(pytree)
    return [
        pytree._pytree__meta.get(field, {})
        for field, _ in sorted(vars(pytree).items())
        if field not in pytree._pytree__static_fields
    ]


def meta_leaves(
    pytree: Module,
    *,
    is_leaf: Optional[Callable[[Any], bool]] = None,
) -> List[Tuple[Optional[Dict[str, Any]], Any]]:
    """
    Returns the meta of the leaves of the pytree.

    Args:
        pytree (Module): pytree to get the meta of.
        is_leaf (Callable[[Any], bool]): predicate to determine if a node is a leaf. Defaults to None.

    Returns
    -------
        List[Tuple[Dict[str, Any], Any]]: meta of the leaves of the pytree.
    """

    def _unpack_metadata(
        meta_leaf: Any,
        pytree: Union[Module, Any],
        is_leaf: Optional[Callable[[Any], bool]],
    ):
        """Recursively unpack leaf metadata."""
        if is_leaf and is_leaf(pytree):
            yield meta_leaf
            return

        if type(pytree) in _registry:  # Registry tree trick, thanks to PyTreeClass!
            leaves_values, _ = _registry[type(pytree)].to_iter(pytree)
            leaves_meta = _toplevel_meta(pytree)

        elif pytree is not None:
            yield meta_leaf
            return

        for metadata, leaf in zip(leaves_meta, leaves_values):
            yield from _unpack_metadata((metadata, leaf), leaf, is_leaf)

    return list(_unpack_metadata(pytree, pytree, is_leaf))


def meta_flatten(
    pytree: Union[Module, Any], *, is_leaf: Optional[Callable[[Any], bool]] = None
) -> Union[Module, Any]:
    """
    Returns the meta of the Module.

    Args:
        pytree (Module): Module to get the meta of.
        is_leaf (Callable[[Any], bool]): predicate to determine if a node is a leaf. Defaults to None.

    Returns
    -------
        Module: meta of the Module.
    """
    return meta_leaves(pytree, is_leaf=is_leaf), jtu.tree_structure(
        pytree, is_leaf=is_leaf
    )


def meta_map(
    f: Callable[[Any, Dict[str, Any]], Any],
    pytree: Union[Module, Any],
    *rest: Any,
    is_leaf: Optional[Callable[[Any], bool]] = None,
) -> Union[Module, Any]:
    """Apply a function to a Module where the first argument are the pytree leaves, and the second argument are the Module metadata leaves.
    Args:
        f (Callable[[Any, Dict[str, Any]], Any]): The function to apply to the pytree.
        pytree (Module): The pytree to apply the function to.
        rest (Any, optional): Additional pytrees to apply the function to. Defaults to None.
        is_leaf (Callable[[Any], bool], optional): predicate to determine if a node is a leaf. Defaults to None.

    Returns
    -------
        Module: The transformed pytree.
    """
    leaves, treedef = meta_flatten(pytree, is_leaf=is_leaf)
    all_leaves = [leaves] + [treedef.treedef.flatten_up_to(r) for r in rest]
    return treedef.unflatten(f(*xs) for xs in zip(*all_leaves))


def meta(pytree: Module, *, is_leaf: Optional[Callable[[Any], bool]] = None) -> Module:
    """Returns the metadata of the Module as a pytree.

    Args:
        pytree (Module): pytree to get the metadata of.

    Returns
    -------
        Module: metadata of the pytree.
    """

    def _filter_meta(meta_leaf):
        meta, _ = meta_leaf
        return meta

    return meta_map(_filter_meta, pytree, is_leaf=is_leaf)


# Model saving and loading. Based upon the Flax checkpointing code
# https://github.com/google/flax/blob/main/flax/training/checkpoints.py
def _is_multiprocess_array(value: Any) -> bool:
    if isinstance(value, jax.Array):
        return not value.is_fully_addressable
    return False


def save_tree(
    path: str, model: Module, overwrite: bool = False, iterate: int = None
) -> None:
    def save_args_from_target(target: Any) -> Any:
        return jax.tree_util.tree_map(
            lambda x: SaveArgs(aggregate=not _is_multiprocess_array(x)), target
        )

    # Include the optimiser's iterate to the checkpoint path.
    if iterate:
        path = os.path.join(path, f"step_{iterate}")

    # Extract the leaves from the model.
    save_args = save_args_from_target(model)

    # Save the model.
    orbax_checkpointer = Checkpointer(PyTreeCheckpointHandler())
    orbax_checkpointer.save(path, model, save_args=save_args, force=overwrite)


def load_tree(path: str, model: Module) -> Module:
    def make_restore_args(x):
        if _is_multiprocess_array(x):
            return ArrayRestoreArgs(
                restore_type=jax.Array,
                sharding=x.sharding,
            )
        return RestoreArgs()

    restore_args = jax.tree_util.tree_map(make_restore_args, model)
    orbax_checkpointer = Checkpointer(PyTreeCheckpointHandler())
    restored = orbax_checkpointer.restore(path, item=model, restore_args=restore_args)
    return restored



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['gpjax',
 'gpjax.base',
 'gpjax.kernels',
 'gpjax.kernels.approximations',
 'gpjax.kernels.computations',
 'gpjax.kernels.non_euclidean',
 'gpjax.kernels.nonstationary',
 'gpjax.kernels.stationary',
 'gpjax.linops']

package_data = \
{'': ['*']}

install_requires = \
['beartype>=0.13.1,<0.14.0',
 'jax>=0.4.1',
 'jaxlib==0.4.7',
 'jaxtyping>=0.2.15,<0.3.0',
 'optax>=0.1.4,<0.2.0',
 'orbax-checkpoint>=0.2.0,<0.3.0',
 'plum-dispatch>=2.1.0,<3.0.0',
 'simple-pytree>=0.1.7,<0.2.0',
 'tensorflow-probability>=0.19.0,<0.20.0',
 'tqdm>=4.65.0,<5.0.0']

setup_kwargs = {
    'name': 'gpjax',
    'version': '0.6.5',
    'description': 'Gaussian processes in JAX.',
    'long_description': '<!-- <h1 align=\'center\'>GPJax</h1>\n<h2 align=\'center\'>Gaussian processes in Jax.</h2> -->\n<p align="center">\n<img width="700" height="300" src="https://raw.githubusercontent.com/JaxGaussianProcesses/GPJax/main/docs/_static/gpjax_logo.svg" alt="GPJax\'s logo">\n</p>\n\n[![codecov](https://codecov.io/gh/JaxGaussianProcesses/GPJax/branch/master/graph/badge.svg?token=DM1DRDASU2)](https://codecov.io/gh/JaxGaussianProcesses/GPJax)\n[![CodeFactor](https://www.codefactor.io/repository/github/jaxgaussianprocesses/gpjax/badge)](https://www.codefactor.io/repository/github/jaxgaussianprocesses/gpjax)\n[![Netlify Status](https://api.netlify.com/api/v1/badges/d3950e6f-321f-4508-9e52-426b5dae2715/deploy-status)](https://app.netlify.com/sites/endearing-crepe-c2d5fe/deploys)\n[![PyPI version](https://badge.fury.io/py/GPJax.svg)](https://badge.fury.io/py/GPJax)\n[![DOI](https://joss.theoj.org/papers/10.21105/joss.04455/status.svg)](https://doi.org/10.21105/joss.04455)\n[![Downloads](https://pepy.tech/badge/gpjax)](https://pepy.tech/project/gpjax)\n[![Slack Invite](https://img.shields.io/badge/Slack_Invite--blue?style=social&logo=slack)](https://join.slack.com/t/gpjax/shared_invite/zt-1da57pmjn-rdBCVg9kApirEEn2E5Q2Zw)\n\n[**Quickstart**](#simple-example)\n| [**Install guide**](#installation)\n| [**Documentation**](https://docs.jaxgaussianprocesses.com/)\n| [**Slack Community**](https://join.slack.com/t/gpjax/shared_invite/zt-1da57pmjn-rdBCVg9kApirEEn2E5Q2Zw)\n\nGPJax aims to provide a low-level interface to Gaussian process (GP) models in\n[Jax](https://github.com/google/jax), structured to give researchers maximum\nflexibility in extending the code to suit their own needs. The idea is that the\ncode should be as close as possible to the maths we write on paper when working\nwith GP models.\n\n# Package support\n\nGPJax was founded by [Thomas Pinder](https://github.com/thomaspinder). Today,\nthe maintenance of GPJax is undertaken by [Thomas\nPinder](https://github.com/thomaspinder) and [Daniel\nDodd](https://github.com/Daniel-Dodd).\n\nWe would be delighted to receive contributions from interested individuals and\ngroups. To learn how you can get involved, please read our [guide for\ncontributing](https://github.com/JaxGaussianProcesses/GPJax/blob/master/CONTRIBUTING.md).\nIf you have any questions, we encourage you to [open an\nissue](https://github.com/JaxGaussianProcesses/GPJax/issues/new/choose). For\nbroader conversations, such as best GP fitting practices or questions about the\nmathematics of GPs, we invite you to [open a\ndiscussion](https://github.com/JaxGaussianProcesses/GPJax/discussions).\n\nFeel free to join our [Slack\nChannel](https://join.slack.com/t/gpjax/shared_invite/zt-1da57pmjn-rdBCVg9kApirEEn2E5Q2Zw),\nwhere we can discuss the development of GPJax and broader support for Gaussian\nprocess modelling.\n\n# Supported methods and interfaces\n\n## Notebook examples\n\n> - [**Conjugate Inference**](https://docs.jaxgaussianprocesses.com/examples/regression/)\n> - [**Classification with MCMC**](https://docs.jaxgaussianprocesses.com/examples/classification/)\n> - [**Sparse Variational Inference**](https://docs.jaxgaussianprocesses.com/examples/collapsed_vi/)\n> - [**Stochastic Variational Inference**](https://docs.jaxgaussianprocesses.com/examples/uncollapsed_vi/)\n> - [**BlackJax Integration**](https://docs.jaxgaussianprocesses.com/examples/classification/#mcmc-inference)\n> - [**Laplace Approximation**](https://docs.jaxgaussianprocesses.com/examples/classification/#laplace-approximation)\n> - [**Inference on Non-Euclidean Spaces**](https://docs.jaxgaussianprocesses.com/examples/kernels/#custom-kernel)\n> - [**Inference on Graphs**](https://docs.jaxgaussianprocesses.com/examples/graph_kernels/)\n> - [**Pathwise Sampling**](https://docs.jaxgaussianprocesses.com/examples/spatial/)\n> - [**Learning Gaussian Process Barycentres**](https://docs.jaxgaussianprocesses.com/examples/barycentres/)\n> - [**Deep Kernel Regression**](https://docs.jaxgaussianprocesses.com/examples/deep_kernels/)\n> - [**Poisson Regression**](https://docs.jaxgaussianprocesses.com/examples/poisson/)\n\n## Guides for customisation\n>\n> - [**Custom kernels**](https://docs.jaxgaussianprocesses.com/examples/kernels/#custom-kernel)\n> - [**UCI regression**](https://docs.jaxgaussianprocesses.com/examples/yacht/)\n\n## Conversion between `.ipynb` and `.py`\nAbove examples are stored in [examples](examples) directory in the double\npercent (`py:percent`) format. Checkout [jupytext\nusing-cli](https://jupytext.readthedocs.io/en/latest/using-cli.html) for more\ninfo.\n\n* To convert `example.py` to `example.ipynb`, run:\n\n```bash\njupytext --to notebook example.py\n```\n\n* To convert `example.ipynb` to `example.py`, run:\n\n```bash\njupytext --to py:percent example.ipynb\n```\n\n# Simple example\n\nLet us import some dependencies and simulate a toy dataset $\\mathcal{D}$.\n\n```python\nimport gpjax as gpx\nfrom jax import grad, jit\nimport jax.numpy as jnp\nimport jax.random as jr\nimport optax as ox\n\nkey = jr.PRNGKey(123)\n\nf = lambda x: 10 * jnp.sin(x)\n\nn = 50\nx = jr.uniform(key=key, minval=-3.0, maxval=3.0, shape=(n,1)).sort()\ny = f(x) + jr.normal(key, shape=(n,1))\nD = gpx.Dataset(X=x, y=y)\n\n# Construct the prior\nmeanf = gpx.mean_functions.Zero()\nkernel = gpx.kernels.RBF()\nprior = gpx.Prior(mean_function=meanf, kernel = kernel)\n\n# Define a likelihood\nlikelihood = gpx.Gaussian(num_datapoints = n)\n\n# Construct the posterior\nposterior = prior * likelihood\n\n# Define an optimiser\noptimiser = ox.adam(learning_rate=1e-2)\n\n# Define the marginal log-likelihood\nnegative_mll = jit(gpx.objectives.ConjugateMLL(negative=True))\n\n# Obtain Type 2 MLEs of the hyperparameters\nopt_posterior, history = gpx.fit(\n    model=posterior,\n    objective=negative_mll,\n    train_data=D,\n    optim=optimiser,\n    num_iters=500,\n    safe=True,\n    key=key,\n)\n\n# Infer the predictive posterior distribution\nxtest = jnp.linspace(-3., 3., 100).reshape(-1, 1)\nlatent_dist = opt_posterior(xtest, D)\npredictive_dist = opt_posterior.likelihood(latent_dist)\n\n# Obtain the predictive mean and standard deviation\npred_mean = predictive_dist.mean()\npred_std = predictive_dist.stddev()\n```\n\n# Installation\n\n## Stable version\n\nThe latest stable version of GPJax can be installed via\npip:\n\n```bash\npip install gpjax\n```\n\n> **Note**\n>\n> We recommend you check your installation version:\n> ```python\n> python -c \'import gpjax; print(gpjax.__version__)\'\n> ```\n\n\n\n## Development version\n> **Warning**\n>\n> This version is possibly unstable and may contain bugs.\n\n> **Note**\n>\n> We advise you create virtual environment before installing:\n> ```\n> conda create -n gpjax_experimental python=3.10.0\n> conda activate gpjax_experimental\n>  ```\n\n\nClone a copy of the repository to your local machine and run the setup\nconfiguration in development mode.\n```bash\ngit clone https://github.com/JaxGaussianProcesses/GPJax.git\ncd GPJax\npoetry install\n```\n\n> We recommend you check your installation passes the supplied unit tests:\n>\n> ```python\n> poetry run pytest\n> ```\n\n# Citing GPJax\n\nIf you use GPJax in your research, please cite our [JOSS paper](https://joss.theoj.org/papers/10.21105/joss.04455#).\n\n```\n@article{Pinder2022,\n  doi = {10.21105/joss.04455},\n  url = {https://doi.org/10.21105/joss.04455},\n  year = {2022},\n  publisher = {The Open Journal},\n  volume = {7},\n  number = {75},\n  pages = {4455},\n  author = {Thomas Pinder and Daniel Dodd},\n  title = {GPJax: A Gaussian Process Framework in JAX},\n  journal = {Journal of Open Source Software}\n}\n```\n',
    'author': 'Thomas Pinder',
    'author_email': 'tompinder@live.co.uk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/JaxGaussianProcesses/GPJax',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.12',
}


setup(**setup_kwargs)



# Copyright 2022 The JaxLinOp Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from gpjax.linops.constant_diagonal_linear_operator import (
    ConstantDiagonalLinearOperator,
)
from gpjax.linops.dense_linear_operator import DenseLinearOperator
from gpjax.linops.diagonal_linear_operator import DiagonalLinearOperator
from gpjax.linops.identity_linear_operator import IdentityLinearOperator
from gpjax.linops.linear_operator import LinearOperator
from gpjax.linops.triangular_linear_operator import (
    LowerTriangularLinearOperator,
    UpperTriangularLinearOperator,
)
from gpjax.linops.utils import (
    identity,
    to_dense,
)
from gpjax.linops.zero_linear_operator import ZeroLinearOperator

__all__ = [
    "LinearOperator",
    "DenseLinearOperator",
    "DiagonalLinearOperator",
    "ConstantDiagonalLinearOperator",
    "IdentityLinearOperator",
    "ZeroLinearOperator",
    "LowerTriangularLinearOperator",
    "UpperTriangularLinearOperator",
    "identity",
    "to_dense",
]



# Copyright 2022 The JaxGaussianProcesses Contributors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================


from dataclasses import dataclass

import jax.numpy as jnp
from jaxtyping import (
    Float,
    Int,
    Num,
)
import tensorflow_probability.substrates.jax as tfp

from gpjax.base import (
    param_field,
    static_field,
)
from gpjax.kernels.base import AbstractKernel
from gpjax.kernels.computations import (
    AbstractKernelComputation,
    EigenKernelComputation,
)
from gpjax.kernels.non_euclidean.utils import jax_gather_nd
from gpjax.typing import (
    Array,
    ScalarFloat,
    ScalarInt,
)

tfb = tfp.bijectors


##########################################
# Graph kernels
##########################################
@dataclass
class GraphKernel(AbstractKernel):
    r"""The Matérn graph kernel defined on the vertex set of a graph.

    A Matérn graph kernel defined on the vertices of a graph. The key reference
    for this object is borovitskiy et. al., (2020).

    Args:
        laplacian (Float[Array]): An $`N \times N`$ matrix representing the Laplacian matrix
            of a graph.
    """

    laplacian: Num[Array, "N N"] = static_field(None)
    lengthscale: ScalarFloat = param_field(jnp.array(1.0), bijector=tfb.Softplus())
    variance: ScalarFloat = param_field(jnp.array(1.0), bijector=tfb.Softplus())
    smoothness: ScalarFloat = param_field(jnp.array(1.0), bijector=tfb.Softplus())
    eigenvalues: Float[Array, " N"] = static_field(None)
    eigenvectors: Float[Array, "N N"] = static_field(None)
    num_vertex: ScalarInt = static_field(None)
    compute_engine: AbstractKernelComputation = static_field(EigenKernelComputation)
    name: str = "Graph Matérn"

    def __post_init__(self):
        if self.laplacian is None:
            raise ValueError("Graph laplacian must be specified")

        evals, self.eigenvectors = jnp.linalg.eigh(self.laplacian)
        self.eigenvalues = evals.reshape(-1, 1)
        if self.num_vertex is None:
            self.num_vertex = self.eigenvalues.shape[0]

    def __call__(  # TODO not consistent with general kernel interface
        self,
        x: Int[Array, "N 1"],
        y: Int[Array, "N 1"],
        *,
        S,
        **kwargs,
    ):
        r"""Compute the (co)variance between a vertex pair.

        For a graph $`\mathcal{G} = \{V, E\}`$ where $`V = \{v_1, v_2, \ldots v_n \}`$,
        evaluate the graph kernel on a pair of vertices $`(v_i, v_j)`$ for any $`i,j<n`$.

        Args:
            x (Float[Array, "N 1"]): Index of the $`i`$th vertex.
            y (Float[Array, "N 1"]): Index of the $`j`$th vertex.

        Returns
        -------
            ScalarFloat: The value of $k(v_i, v_j)$.
        """
        Kxx = (jax_gather_nd(self.eigenvectors, x) * S.squeeze()) @ jnp.transpose(
            jax_gather_nd(self.eigenvectors, y)
        )  # shape (n,n)
        return Kxx.squeeze()
